""" Products.mcdutils session data container """
import six

from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo
from OFS.PropertyManager import PropertyManager
from OFS.SimpleItem import SimpleItem
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from zope.interface import implementedBy
from zope.interface import implementer

from .interfaces import IMemCacheSessionDataContainer
from .mapping import MemCacheMapping


@implementer(IMemCacheSessionDataContainer + implementedBy(SimpleItem)
             + implementedBy(PropertyManager))
class MemCacheSessionDataContainer(SimpleItem, PropertyManager):
    """ Implement ISDC via a memcache proxy.
    """
    security = ClassSecurityInfo()

    _v_proxy = None
    proxy_path = ''
    zmi_icon = 'far fa-clock'

    def __init__(self, id, title=''):
        self.id = id
        self.title = title

    def _get_proxy(self):
        if self._v_proxy is None:
            if not self.proxy_path:
                from Products.mcdutils import MemCacheError
                raise MemCacheError('No proxy defined')
            self._v_proxy = self.unrestrictedTraverse(self.proxy_path)
        return self._v_proxy

    # proxy = property(_get_proxy,)  # can't acquire inside property!

    #
    #   ZMI
    #
    meta_type = 'MemCache Session Data Container'
    _properties = (
        {'id': 'proxy_path', 'type': 'string', 'mode': 'w'},
    )

    manage_options = (
        PropertyManager.manage_options
        + ({'action': 'addItemsToSessionForm', 'label': 'Test'},)
        + SimpleItem.manage_options)

    security.declarePublic('addItemsToSessionForm')  # NOQA: D001
    addItemsToSessionForm = PageTemplateFile('www/add_items.pt', globals())

    security.declarePublic('addItemsToSession')  # NOQA: D001

    def addItemsToSession(self):
        """ Add key value pairs from 'items' textarea to the session.
        """
        request = self.REQUEST
        items = request.form.get('items', ())
        session = request['SESSION']

        before = len(session)
        count = len(items)

        for line in items:
            k, v = line.split(b' ', 1)
            k = k.strip()
            v = v.strip()
            session[k] = v

        after = len(session)

        return 'Before: %d;  after: %d; # items: %d' % (before, after, count)

    #
    #   ISessionDataContainer implementation
    #
    security.declarePrivate('has_key')  # NOQA: D001

    def has_key(self, key):
        """ See ISessionDataContainer.
        """
        return self._get_proxy().get(self._safe_key(key)) is not None

    security.declarePrivate('new_or_existing')  # NOQA: D001

    def new_or_existing(self, key):
        """ See ISessionDataContainer.
        """
        key = self._safe_key(key)
        mapping = self.get(key)

        if mapping is None:
            proxy = self._get_proxy()
            mapping = MemCacheMapping(key, proxy)
            proxy._cached[key] = mapping

        return mapping

    security.declarePrivate('get')  # NOQA: D001

    def get(self, key):
        """ See ISessionDataContainer.
        """
        return self._get_proxy().get(self._safe_key(key))

    def _safe_key(self, key):
        """ Helper to ensure the key is always a binary string """
        if isinstance(key, six.text_type):
            key = key.encode('UTF-8')
        return key


InitializeClass(MemCacheSessionDataContainer)


def addMemCacheSessionDataContainer(dispatcher, id, title='', REQUEST=None):
    """ Add a MCSDC to dispatcher.
    """
    dispatcher._setObject(id, MemCacheSessionDataContainer(id, title=title))

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect('%s/manage_workspace'
                                     % dispatcher.absolute_url())


addMemCacheSessionDataContainerForm = PageTemplateFile('www/add_mcsdc.pt',
                                                       globals())



""" memcache-aware transactional mapping """
import transaction
from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo
from persistent.mapping import PersistentMapping
from transaction.interfaces import IDataManagerSavepoint
from transaction.interfaces import ISavepointDataManager
from zope.interface import implementedBy
from zope.interface import implementer


@implementer(ISavepointDataManager + implementedBy(PersistentMapping))
class MemCacheMapping(PersistentMapping):
    """ memcache-based mapping which manages its own transactional semantics
    """
    security = ClassSecurityInfo()

    def __init__(self, key, proxy):
        PersistentMapping.__init__(self)
        self._p_oid = hash(key)
        self._p_jar = self   # we are our own data manager
        self._p_key = key
        self._p_proxy = proxy
        self._p_joined = False

    security.setDefaultAccess('allow')
    security.declareObjectPublic()

    set = PersistentMapping.__setitem__
    __guarded_setitem__ = PersistentMapping.__setitem__
    __guarded_delitem__ = PersistentMapping.__delitem__
    delete = PersistentMapping.__delitem__

    def __of__(self, other):
        # Behavior change from prior versions due to issues in Python 3:
        # This class used to subclass from Acquisition.Explicit as well,
        # but under Python 3 that causes metaclass conflicts since
        # PersistentMapping's metaclass is abc.ABCMeta and
        # Acquisition.Explicit has ExtensionClass as metaclass.
        # Nasty workaround: Stub out ``__of__`` as the session data manager
        # still expects us to support ``__of__``, but this object does
        # _not_ support Acquisition anymore.
        return self

    def __getstate__(self):
        return self.data

    def __setstate__(self, value):
        self.data = {}
        self.data.update(value)

    def __repr__(self):
        # Overriding here to try and hide some password fields, like
        # the ZPublisher HTTPRequest class tries to do.
        new_dict = dict(self.data)
        for key in new_dict.keys():
            if 'passw' in key.lower():
                new_dict[key] = '<password obscured>'
        return repr(new_dict)

    def has_key(self, key):
        """ Backwards compatibility under Python 3 """
        return key in self.data

    def getContainerKey(self):
        """ Fake out (I)Transient API.
        """
        return self._p_key

    def _clean(self):
        # Remove from proxy cache to force an update
        # from memcached during next access.
        try:
            del self._p_proxy._cached[self._p_key]
        except KeyError:
            pass

    security.declarePrivate('abort')  # NOQA: D001

    def abort(self, txn):
        """ See IDataManager.
        """
        self._clean()

    security.declarePrivate('tpc_begin')  # NOQA: D001

    def tpc_begin(self, txn):
        """ See IDataManager.
        """

    security.declarePrivate('commit')  # NOQA: D001

    def commit(self, txn):
        """ See IDataManager.
        """

    security.declarePrivate('tpc_vote')  # NOQA: D001

    def tpc_vote(self, txn):
        """ See IDataManager.
        """
        server, key = self._p_proxy.client._get_server(self._p_key)
        if server is None:
            from Products.mcdutils import MemCacheError
            raise MemCacheError("Can't reach memcache server!")

    security.declarePrivate('tpc_finish')  # NOQA: D001

    def tpc_finish(self, txn):
        """ See IDataManager.
        """
        if self._p_changed:
            self._p_proxy.set(self._p_key, self)  # no error handling
        self._p_changed = 0
        self._p_joined = False
        self._clean()

    security.declarePrivate('tpc_abort')  # NOQA: D001

    def tpc_abort(self, txn):
        """ See IDataManager.
        """
        self._p_joined = False
        self._p_changed = 0
        self._clean()

    security.declarePrivate('sortKey')  # NOQA: D001

    def sortKey(self):
        """ See IDataManager.
        """
        return 'MemCacheMapping: %s' % self._p_key

    security.declarePrivate('register')  # NOQA: D001

    def register(self, obj):
        """ See IPersistentDataManager
        """
        if obj is not self:
            raise ValueError("Can't be the jar for another object.")

        if not self._p_joined:
            transaction.get().join(self)
            self._p_joined = True

    security.declarePrivate('savepoint')  # NOQA: D001

    def savepoint(self):
        """ See ITransaction
        """
        return MemCacheMappingSavepoint()


InitializeClass(MemCacheMapping)


@implementer(IDataManagerSavepoint)
class MemCacheMappingSavepoint(object):
    """ A simple savepoint object
    """

    def rollback(self):
        """ Roll back a savepoint

        Memcache and the python-memcached library don't have the concept
        of a rollback, so there is nothing useful to do here.
        """
        pass



""" Unit tests for Products.mcdutils.mapping """
import unittest


class MemCacheMappingSavepointTests(unittest.TestCase):

    def _getTargetClass(self):
        from ..mapping import MemCacheMappingSavepoint
        return MemCacheMappingSavepoint

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()()

    def test_conforms_to_IDataManagerSavepoint(self):
        from transaction.interfaces import IDataManagerSavepoint
        from zope.interface.verify import verifyClass
        verifyClass(IDataManagerSavepoint, self._getTargetClass())

    def test_rollback(self):
        # This doesn't really do anything. Just verifying the
        # method is there and doesn't blow up when called.
        sp = self._makeOne()
        self.assertFalse(sp.rollback())


class MemCacheMappingTests(unittest.TestCase):

    def _getTargetClass(self):
        from Products.mcdutils.mapping import MemCacheMapping
        return MemCacheMapping

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def test_conforms_to_ISavepointDataManager(self):
        from transaction.interfaces import ISavepointDataManager
        from zope.interface.verify import verifyClass
        verifyClass(ISavepointDataManager, self._getTargetClass())

    def test___setitem___triggers_register(self):
        mapping = self._makeOne('key', DummyProxy())
        self.assertFalse(mapping._p_changed)
        self.assertFalse(mapping._p_joined)
        mapping['abc'] = 123
        self.assertTrue(mapping._p_changed)
        self.assertTrue(mapping._p_joined)

    def test_has_key(self):
        # Added in for backwards-compatibility under Python 3
        mapping = self._makeOne('key', DummyProxy())

        self.assertFalse(mapping.has_key('foo'))  # NOQA: W601
        mapping['foo'] = 'bar'
        self.assertTrue(mapping.has_key('foo'))  # NOQA: W601

    def test__getstate__and__setstate__(self):
        mapping = self._makeOne('key', DummyProxy())

        self.assertEqual(mapping.__getstate__(), {})
        mapping.__setstate__({'foo': 'bar'})
        self.assertEqual(mapping.__getstate__(), {'foo': 'bar'})

    def test_getContainerKey(self):
        mapping = self._makeOne('key', DummyProxy())

        self.assertEqual(mapping.getContainerKey(), 'key')

    def test_clean(self):
        proxy = DummyProxy()
        proxy._set('key', 'myvalue')
        mapping = self._makeOne('key', proxy)

        self.assertIn('key', proxy._cached)
        mapping._clean()
        self.assertNotIn('key', proxy._cached)

        # Cleaning again won't throw errors
        self.assertIsNone(mapping._clean())

    def test_abort(self):
        proxy = DummyProxy()
        proxy._set('key', 'myvalue')
        mapping = self._makeOne('key', proxy)

        self.assertIn('key', proxy._cached)
        mapping.abort(None)
        self.assertNotIn('key', proxy._cached)

    def test_savepoint(self):
        from Products.mcdutils.mapping import MemCacheMappingSavepoint
        mapping = self._makeOne('key', DummyProxy())

        sp = mapping.savepoint()
        self.assertIsInstance(sp, MemCacheMappingSavepoint)

    def test_sortKey(self):
        mapping = self._makeOne('key', DummyProxy())

        self.assertEqual(mapping.sortKey(), 'MemCacheMapping: key')

    def test_repr(self):
        KEYS = ('__ac_password', 'passwd', 'password')
        proxy = DummyProxy()
        proxy._set('key', 'myvalue')
        mapping = self._makeOne('key', proxy)

        for pw_key in KEYS:
            mapping[pw_key] = 'thisisapw'
        mapping['normal'] = 'normalvalue'

        mapping_repr = repr(mapping)
        self.assertNotIn('thisisapw', mapping_repr)
        for pw_key in KEYS:
            self.assertIn("'%s': '<password obscured>'" % pw_key, mapping_repr)
        self.assertIn("'normal': 'normalvalue'", mapping_repr)


class DummyClient:
    def _get_server(self, key):
        return self, key


class DummyProxy:

    def __init__(self):
        self._cached = {}

    def _set(self, key, value):
        self._cached[key] = value

    def _clean(self, key):
        try:
            del self._cached[key]
        except KeyError:
            pass

    client = DummyClient()


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(MemCacheMappingTests))
    suite.addTest(unittest.makeSuite(MemCacheMappingSavepointTests))
    return suite



""" Product:  mcdutils

Implement Zope sessions using memcached as the backing store.
"""


class MemCacheError(IOError):
    pass


def initialize(context):

    from .proxy import MemCacheProxy
    from .proxy import addMemCacheProxy
    from .proxy import addMemCacheProxyForm
    context.registerClass(MemCacheProxy,
                          constructors=(addMemCacheProxyForm,
                                        addMemCacheProxy),
                          icon='www/proxy.gif')

    from .sessiondata import MemCacheSessionDataContainer
    from .sessiondata import addMemCacheSessionDataContainer
    from .sessiondata import addMemCacheSessionDataContainerForm
    context.registerClass(MemCacheSessionDataContainer,
                          constructors=(addMemCacheSessionDataContainerForm,
                                        addMemCacheSessionDataContainer),
                          icon='www/sdc.gif')

    from .zcache import MemCacheZCacheManager
    from .zcache import addMemCacheZCacheManager
    from .zcache import addMemCacheZCacheManagerForm
    context.registerClass(MemCacheZCacheManager,
                          constructors=(addMemCacheZCacheManagerForm,
                                        addMemCacheZCacheManager),
                          icon='www/zcm.gif')



""" RAMCacheManager workalike using memcache """
from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo
from OFS.Cache import CacheManager
from OFS.PropertyManager import PropertyManager
from OFS.SimpleItem import SimpleItem
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from zope.interface import implementedBy
from zope.interface import implementer

from .interfaces import IZCache
from .interfaces import IZCacheManager


def aggregateKey(ob, view_name='', request=None, request_names=(),
                 local_keys=None):
    """ Return a key to be used when retrieving or inserting a cache entry.

    o 'ob' is the object for whom the key is desired.

    o 'view_name' must be a string.

    o 'request' must be a mapping.

    o 'request_names' must be a sorted list of keys to be included in the key
      from 'request'

    o 'local_keys' is a mapping or None.
    """
    path = '/'.join(ob.getPhysicalPath())
    request_index = []
    local_index = []
    if request is None:
        request = {}
    if local_keys is None:
        local_keys = {}

    for key in request_names:
        val = request.get(key, '')
        request_index.append('%s:%s' % (str(key), str(val)))

    for key, val in local_keys.items():
        local_index.append('%s:%s' % (str(key), str(val)))

    full_key = '|'.join((path, str(view_name),
                         ','.join(request_index),
                         ','.join(sorted(local_index))))

    # Memcache does not like  blank spaces in keys
    return full_key.replace(' ', '_')


@implementer(IZCache)
class MemCacheZCache(object):
    """ Implement ISDC via a memcache proxy.
    """
    security = ClassSecurityInfo()
    security.declareObjectPrivate()

    def __init__(self, proxy, request_names):
        self.proxy = proxy
        self.request_names = request_names

    def ZCache_invalidate(self, ob):
        """ See IZCache.
        """
        path = '/'.join(ob.getPhysicalPath())
        proxy = self.proxy
        keys = proxy.get(path)
        if keys is None:
            return
        for key in keys:
            proxy.delete(key)
        proxy.delete(path)

    def ZCache_get(self, ob, view_name='', keywords=None, mtime_func=None,
                   default=None):
        """ See IZCache.
        """
        key = self._getKey(ob, view_name, keywords)

        value = self.proxy.get(key)

        if value is None:
            return default

        return value

    def ZCache_set(self, ob, data, view_name='', keywords=None,
                   mtime_func=None):
        """ See IZCache.
        """
        path = '/'.join(ob.getPhysicalPath())
        proxy = self.proxy
        key = self._getKey(ob, view_name, keywords)
        proxy.set(key, data)

        keys = proxy.get(path)
        if keys is None:
            keys = {}
        if key not in keys:
            keys[key] = 1
            proxy.set(path, keys)

    def _getKey(self, ob, view_name, keywords):
        rnames = self.request_names
        if rnames:
            request = getattr(ob, 'REQUEST', {})
        else:
            request = {}

        return aggregateKey(ob, view_name, request, rnames, keywords)


InitializeClass(MemCacheZCache)


@implementer(IZCacheManager + implementedBy(CacheManager)
             + implementedBy(SimpleItem) + implementedBy(PropertyManager))
class MemCacheZCacheManager(CacheManager, SimpleItem, PropertyManager):
    """ Implement ISDC via a memcache proxy.
    """
    security = ClassSecurityInfo()

    _v_proxy = None
    proxy_path = ''
    request_names = ()
    zmi_icon = 'fas fa-forward'

    #
    #   ZMI
    #
    meta_type = 'MemCache Cache Manager'
    _properties = (
        {'id': 'title', 'type': 'string', 'mode': 'w'},
        {'id': 'proxy_path', 'type': 'string', 'mode': 'w'},
        {'id': 'request_names', 'type': 'lines', 'mode': 'w'},
    )

    manage_options = (PropertyManager.manage_options
                      + CacheManager.manage_options
                      + SimpleItem.manage_options)

    def __init__(self, id, title=''):
        self.id = id
        self.title = title

    def _get_proxy(self):
        if self._v_proxy is None:
            if not self.proxy_path:
                # import late to avoid cycle
                from . import MemCacheError
                raise MemCacheError('No proxy defined')
            self._v_proxy = self.unrestrictedTraverse(self.proxy_path)
        return self._v_proxy

    def ZCacheManager_getCache(self):
        """ See IZCacheManager.
        """
        names = list(self.request_names)
        names.sort()
        return MemCacheZCache(self._get_proxy(), tuple(names))


InitializeClass(MemCacheZCacheManager)


def addMemCacheZCacheManager(dispatcher, id, title='', REQUEST=None):
    """ Add a MCSDC to dispatcher.
    """
    dispatcher._setObject(id, MemCacheZCacheManager(id, title=title))

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect('%s/manage_workspace'
                                     % dispatcher.absolute_url())


addMemCacheZCacheManagerForm = PageTemplateFile('www/add_mczcm.pt', globals())



""" Unit tests for Products.mcdutils.proxy """
import unittest


KEY = b'key1'


class FauxClientTests(unittest.TestCase):

    def _getTargetClass(self):
        from ..proxy import FauxClient
        return FauxClient

    def _makeOne(self):
        return self._getTargetClass()()

    def test_faux_client(self):
        # Faux client only fakes out a few methods
        fc = self._makeOne()

        self.assertEqual(fc._get_server(KEY), (fc, KEY))
        fc.set(KEY, 'value1')
        self.assertEqual(fc._get_server(KEY), (fc, KEY))


class MemCacheProxyTests(unittest.TestCase):

    def _getTargetClass(self):
        from ..proxy import MemCacheProxy
        return MemCacheProxy

    def _makeOne(self, *args, **kw):
        return self._getTargetClass()(*args, **kw)

    def _makeOneWithMemcache(self, *args, **kw):
        from .helpers import DummyMemcache
        proxy = self._getTargetClass()(*args, **kw)
        proxy._v_client = DummyMemcache()
        return proxy

    def test_conforms_to_IMemCacheProxy(self):
        from zope.interface.verify import verifyClass

        from ..interfaces import IMemCacheProxy
        verifyClass(IMemCacheProxy, self._getTargetClass())

    def test__init__(self):
        proxy = self._makeOne('proxy', title='Proxy')

        self.assertEqual(proxy.getId(), 'proxy')
        self.assertEqual(proxy.servers, ())
        self.assertEqual(proxy.getProperty('servers'), ())
        self.assertEqual(proxy.title, 'Proxy')
        self.assertEqual(proxy.getProperty('title'), 'Proxy')

    def test__cached(self):
        proxy = self._makeOne('proxy')

        self.assertEqual(proxy._cached, {})

        proxy._v_cached = {'foo': 'bar'}
        self.assertEqual(proxy._cached, {'foo': 'bar'})

    def test_client(self):
        from memcache import Client
        proxy = self._makeOne('proxy')

        self.assertIsNotNone(proxy.client)

        proxy._v_client = 'x'
        self.assertEqual(proxy.client, 'x')

        # Set a server, which should create a real client instance
        proxy.servers = ('127.0.0.1:9999',)
        self.assertIsInstance(proxy.client, Client)

    def test__servers(self):
        proxy = self._makeOne('proxy')

        self.assertEqual(proxy.servers, ())
        proxy.servers = ('srv',)
        self.assertEqual(proxy.servers, ('srv',))

        # make sure all caches are cleared
        proxy._v_client = 'client'
        proxy._v_cache = 'cache'
        self.assertIsNotNone(getattr(proxy, '_v_client'))
        self.assertIsNotNone(getattr(proxy, '_v_cache'))
        proxy.servers = ('srv',)
        self.assertIsNone(getattr(proxy, '_v_client', None))
        self.assertIsNone(getattr(proxy, '_v_cache', None))

    def test_create(self):
        from ..mapping import MemCacheMapping
        proxy = self._makeOne('proxy')

        created = proxy.create(KEY)
        self.assertIsInstance(created, MemCacheMapping)

    def test_get_set(self):
        proxy = self._makeOneWithMemcache('proxy')

        self.assertIsNone(proxy.get(KEY))
        self.assertTrue(proxy.set(KEY, proxy.create(KEY)))
        self.assertEqual(proxy.get(KEY), {})

        # This should also work when setting values that are
        # not MemCacheMapping instances
        KEY2 = b'key2'
        self.assertIsNone(proxy.get(KEY2))
        self.assertTrue(proxy.set(KEY2, {'foo': 'bar'}))
        self.assertEqual(proxy.get(KEY2), {'foo': 'bar'})

    def test_get_multi(self):
        proxy = self._makeOneWithMemcache('proxy')

        self.assertEqual(proxy.get_multi([KEY, b'key2']),
                         {KEY: None, b'key2': None})

    def test_add(self):
        proxy = self._makeOneWithMemcache('proxy')

        self.assertTrue(proxy.add(KEY, proxy.create(KEY)))
        self.assertEqual(proxy.get(KEY), {})

    def test_replace(self):
        proxy = self._makeOneWithMemcache('proxy')

        self.assertIsNone(proxy.replace(KEY, proxy.create(KEY)))
        self.assertIsNone(proxy.get(KEY))

        self.assertTrue(proxy.set(KEY, proxy.create(KEY)))
        self.assertEqual(proxy.get(KEY), {})

    def test_delete(self):
        proxy = self._makeOneWithMemcache('proxy')

        self.assertIsNone(proxy.delete(KEY), proxy.create(KEY))

        self.assertTrue(proxy.set(KEY, proxy.create(KEY)))
        self.assertTrue(proxy.delete(KEY))
        self.assertIsNone(proxy.get(KEY))


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(FauxClientTests))
    suite.addTest(unittest.makeSuite(MemCacheProxyTests))
    return suite



""" Memcache proxy """
import memcache

from AccessControl.class_init import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo
from OFS.PropertyManager import PropertyManager
from OFS.SimpleItem import SimpleItem
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from zope.interface import implementedBy
from zope.interface import implementer

from .interfaces import IMemCacheProxy
from .mapping import MemCacheMapping


class FauxClient(dict):

    def _get_server(self, key):
        return self, key

    def set(self, key, value):
        self[key] = value


@implementer(IMemCacheProxy + implementedBy(SimpleItem)
             + implementedBy(PropertyManager))
class MemCacheProxy(SimpleItem, PropertyManager):
    """ Implement ISDC via a a pool of memcache servers.
    """
    security = ClassSecurityInfo()

    _v_cached = None
    _v_client = None
    zmi_icon = 'fas fa-tachometer-alt'

    def __init__(self, id, title=''):
        self.id = id
        self.title = title

    def _get_cached(self):
        if self._v_cached is None:
            self._v_cached = {}
        return self._v_cached

    _cached = property(_get_cached)

    def _get_client(self):
        if self._v_client is not None:
            return self._v_client

        if self.servers:
            client = self._v_client = memcache.Client(self.servers)
        else:
            client = self._v_client = FauxClient()

        return client

    client = property(_get_client)

    _servers = ()

    def _set_servers(self, value):
        self._servers = value
        try:
            del self._v_client
        except AttributeError:
            pass
        try:
            del self._v_cache
        except AttributeError:
            pass

    servers = property(lambda self: self._servers, _set_servers)

    #
    #   ZMI
    #
    meta_type = 'MemCache Proxy'
    _properties = ({'id': 'title', 'type': 'string', 'mode': 'w'},
                   {'id': 'servers', 'type': 'ulines', 'mode': 'w'})

    manage_options = (PropertyManager.manage_options
                      + SimpleItem.manage_options)

    security.declarePrivate('get')  # NOQA: D001

    def get(self, key):
        """ See IMemCacheProxy.
        """
        mapping = self._cached.get(key)

        if mapping is None:
            mapping = self._get_remote(key)

        if isinstance(mapping, MemCacheMapping):
            # Force this mapping to clean up after the transaction.
            mapping.register(mapping)

        return mapping

    security.declarePrivate('get_multi')  # NOQA: D001

    def get_multi(self, keys):
        """ See IMemCacheProxy.
        """
        return self._get_remote_multi(keys)

    security.declarePrivate('set')  # NOQA: D001

    def set(self, key, value):
        """ See IMemCacheProxy.
        """
        rc = self.client.set(key, value)
        try:
            del self._cached[key]
        except KeyError:
            pass
        return rc

    security.declarePrivate('add')  # NOQA: D001

    def add(self, key, value):
        """ Store value (a mapping) for 'key'.

        o Return a boolean to indicate success.

        o Like 'set', but stores value only if the key does not already exist.
        """
        rc = self.client.add(key, value)
        try:
            del self._cached[key]
        except KeyError:
            pass
        return rc

    security.declarePrivate('replace')  # NOQA: D001

    def replace(self, key, value):
        """ Store value (a mapping) for 'key'.

        o Return a boolean to indicate success.

        o Like 'set', but stores value only if the key already exists.
        """
        rc = self.client.replace(key, value)
        try:
            del self._cached[key]
        except KeyError:
            pass
        return rc

    security.declarePrivate('delete')  # NOQA: D001

    def delete(self, key, time=0):
        """ Remove the value stored in the cache under 'key'.

        o Return a boolean to indicate success.

        o 'time', if nonzero, will be tested in the cache to determine
          whether the item's
        """
        rc = self.client.delete(key, time)
        try:
            del self._cached[key]
        except KeyError:
            pass
        return rc

    security.declarePrivate('create')  # NOQA: D001

    def create(self, key):
        """ See IMemCacheProxy.
        """
        mapping = self._cached[key] = MemCacheMapping(key, self)
        return mapping

    #
    #   Helper methods
    #
    def _get_remote(self, key):
        mapping = self._cached[key] = self.client.get(key)
        if isinstance(mapping, MemCacheMapping):
            mapping._p_proxy = self
            mapping._p_key = key
            mapping._p_oid = hash(key)
            mapping._p_jar = mapping
            mapping._p_joined = False
            mapping._p_changed = 0
        return mapping

    def _get_remote_multi(self, keys):
        meta_map = self.client.get_multi(keys)
        result = {}
        for key, mapping in meta_map.items():
            if isinstance(mapping, MemCacheMapping):
                mapping._p_proxy = self
                mapping._p_key = key
                mapping._p_oid = hash(key)
                mapping._p_jar = mapping
                mapping._p_joined = False
                mapping._p_changed = 0
            result[key] = mapping

        return result


InitializeClass(MemCacheProxy)


def addMemCacheProxy(dispatcher, id, title='', REQUEST=None):
    """ Add a MCP to dispatcher.
    """
    dispatcher._setObject(id, MemCacheProxy(id, title=title))

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect('%s/manage_workspace'
                                     % dispatcher.absolute_url())


addMemCacheProxyForm = PageTemplateFile('www/add_mcp.pt', globals())



from setuptools import find_packages
from setuptools import setup


readme = open('README.rst').read()
history = open('CHANGES.txt').read()
long_description = readme + '\n\n' + history

setup(name='Products.mcdutils',
      version='3.0',
      description=('A Zope product with memcached-backed ZCache and '
                   'Zope session implementations.'),
      long_description=long_description,
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Zope',
        'Framework :: Zope :: 4',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP :: Session'],
      keywords='session memcache memcached Products',
      author='Tres Seaver and contributors',
      author_email='tseaver@palladion.com',
      maintainer='Jens Vagelpohl',
      maintainer_email='jens@netz.ooo',
      url='https://mcdutils.readthedocs.io',
      project_urls={
        'Documentation': 'https://mcdutils.readthedocs.io',
        'Issue Tracker': ('https://github.com/dataflake/Products.mcdutils'
                          '/issues'),
        'Sources': 'https://github.com/dataflake/Products.mcdutils',
      },
      license='ZPL 2.1',
      packages=find_packages(),
      include_package_data=True,
      namespace_packages=['Products'],
      zip_safe=False,
      python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
      install_requires=[
        'setuptools',
        'six',
        'python-memcached',
        'Zope >4',
        ],
      extras_require={
        'docs': ['repoze.sphinx.autointerface', 'Sphinx'],
        },
      )



""" Unit test helper modules """
import six


class DummyMemcache(dict):

    def _assertKeyBinary(self, key):
        if not isinstance(key, six.binary_type):
            raise ValueError('Key must be binary string.')
        return key

    def set(self, key, value):
        self._assertKeyBinary(key)
        self[key] = value
        return True

    def add(self, key, value):
        self._assertKeyBinary(key)
        if key not in self:
            self[key] = value
            return True

    def replace(self, key, value):
        self._assertKeyBinary(key)
        if key in self:
            self[key] = value
            return True

    def delete(self, key, time=0):
        self._assertKeyBinary(key)
        if key in self:
            del self[key]
            return True

    def get(self, key):
        self._assertKeyBinary(key)
        if key in self:
            return self[key]

    def get_multi(self, keys):
        res = {}
        for key in keys:
            self._assertKeyBinary(key)
            res[key] = self.get(key)
        return res



""" Unit tests for Products.mcdutils.sessiondata """
import unittest


class DummyClient:
    def _get_server(self, key):
        return self, key


class DummyProxy:
    def __init__(self):
        self._cached = {}

    def set(self, key, value):
        pass

    def _get(self, key, default=None):
        return self._cached.get(key, default)

    get = _get


class MemCacheSessionDataTests(unittest.TestCase):

    def _getTargetClass(self):
        from ..sessiondata import MemCacheSessionDataContainer
        return MemCacheSessionDataContainer

    def _makeOne(self, id, title='', with_proxy=True):
        sdc = self._getTargetClass()(id, title=title)
        if with_proxy:
            sdc.dummy_proxy = DummyProxy()
            sdc.proxy_path = 'dummy_proxy'
        return sdc

    def test_conforms_to_ISessionDataContainer(self):
        from zope.interface.verify import verifyClass

        from ..interfaces import ISessionDataContainer
        verifyClass(ISessionDataContainer, self._getTargetClass())

    def test_conforms_to_IMemCacheSessionDataContainer(self):
        from zope.interface.verify import verifyClass

        from ..interfaces import IMemCacheSessionDataContainer
        verifyClass(IMemCacheSessionDataContainer, self._getTargetClass())

    def test_empty(self):
        sdc = self._makeOne('mcsdc')
        self.assertFalse(sdc.has_key('foobar'))  # NOQA: W601
        self.assertIsNone(sdc.get('foobar'))

    def test_invalid_proxy_raises_MemCacheError(self):
        from .. import MemCacheError
        sdc = self._makeOne('mcsdc', with_proxy=False)
        self.assertRaises(MemCacheError,
                          sdc.has_key, 'foobar')  # NOQA: W601
        self.assertRaises(MemCacheError, sdc.get, 'foobar')
        self.assertRaises(MemCacheError, sdc.new_or_existing, 'foobar')

    def test_new_or_existing_returns_txn_aware_mapping(self):
        from persistent.mapping import PersistentMapping
        from transaction.interfaces import IDataManager
        sdc = self._makeOne('mcsdc')
        created = sdc.new_or_existing('foobar')
        self.assertTrue(isinstance(created, PersistentMapping))
        jar = created._p_jar
        self.assertFalse(jar is None)
        self.assertTrue(IDataManager.providedBy(jar))

    def test_has_key_after_new_or_existing_returns_True(self):
        sdc = self._makeOne('mcsdc')
        sdc.new_or_existing('foobar')
        self.assertTrue(sdc.has_key('foobar'))  # NOQA: W601

    def test_get_after_new_or_existing_returns_same(self):
        sdc = self._makeOne('mcsdc')
        created = sdc.new_or_existing('foobar')
        self.assertTrue(sdc.get('foobar') is created)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(MemCacheSessionDataTests))
    return suite



# Run this test from 'zopectl run'
# Requires that we are running a memcached on localhost, port 11211
import transaction

from .proxy import MemCacheProxy


proxy = MemCacheProxy(['localhost:11211'])

session = proxy.new_or_existing('foobar')
print(session)

session['abc'] = 123
print(session)

transaction.commit()

proxy2 = MemCacheProxy(['localhost:11211'])

print(proxy2.get('foobar'))
